import camb
from . import growth_eqns
from . import solve_eqns
import scipy.integrate
import numpy as np

def compute_pk_gr(params, zf):

    ombh2 = params['Omega_b']*(params['h']/100)**2
    omch2 = params['Omega_m']*(params['h']/100)**2 - ombh2

    pars = camb.CAMBparams()
    pars.set_cosmology(H0=params['h'], ombh2=ombh2, omch2=omch2, mnu=0., tau=0.09)
    pars.InitPower.set_params(As=params['A_s'], ns=params['n_s'])
    pars.set_accuracy(AccuracyBoost=1)
    pars.set_matter_power(redshifts=zf, kmax=20, k_per_logint=0, nonlinear=False)
    #Linear power spectrum
    results = camb.get_results(pars)
    return results.get_linear_matter_power_spectrum()


def compute_pk_fr_vectorized(Omega_m,fR0,n,zvec,kvec,pk_gr):
    """Takes GR matter power spectrum pk_gr at z=0."""

    Nz = len(zvec)
    z_trans = 127
    a_trans = 1/(1+z_trans)
    ai = 1e-10
    growth_lcdm_trans = scipy.integrate.quadrature(growth_eqns.growth_int,ai,a_trans,args=(Omega_m,ai,a_trans),tol=1e-8,rtol=1e-8)[0]
    growth_lcdm_fin = scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,1,args=(Omega_m,ai,1),tol=1e-8,rtol=1e-8)[0]
    D_lcdm = growth_lcdm_trans/growth_lcdm_fin # LCDM growth normalized at a=1
    pk_gr_ini = D_lcdm**2 * pk_gr

    pk_fr = np.repeat(pk_gr_ini, Nz, axis=0) # initialize MG P(k) to GR P(k) at z=z_trans
    for i,k in enumerate(kvec):
        growth_fr = solve_eqns.calc_growth_fr_full(Omega_m,fR0,n,k,0.)
        D_fr = growth_fr(np.log(1/(1+zvec)))[0] * 1/(a_trans*(1+zvec)) # array of Nz values for each k
        pk_fr[:,i] *= D_fr**2
    
    return pk_fr


def compute_pk_dgp_vectorized(Omega_m,rcH0,zvec,pk_gr):
    """Takes GR matter power spectrum pk_gr at z=0."""

    Nz = len(zvec)
    z_trans = 127
    a_trans = 1/(1+z_trans)
    ai = 1e-10
    growth_lcdm_trans = scipy.integrate.quadrature(growth_eqns.growth_int,ai,a_trans,args=(Omega_m,ai,a_trans),tol=1e-8,rtol=1e-8)[0]
    growth_lcdm_fin = scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,1,args=(Omega_m,ai,1),tol=1e-8,rtol=1e-8)[0]
    D_lcdm = growth_lcdm_trans/growth_lcdm_fin # LCDM growth normalized at a=1
    pk_gr_ini = D_lcdm**2 * pk_gr

    pk_dgp = np.repeat(pk_gr_ini, Nz, axis=0) # initialize MG P(k) to GR P(k) at z=z_trans
    growth_dgp = solve_eqns.calc_growth_dgp_full(Omega_m,rcH0,0.)
    D_dgp = growth_dgp(np.log(1/(1+zvec)))[0] * 1/(a_trans*(1+zvec)) # array of Nz values
    D_dgp_mat = np.repeat(D_dgp,len(pk_gr[0])).reshape(pk_dgp.shape) # matrix of Nz * Nk values
    pk_dgp *= D_dgp_mat**2
    
    return pk_dgp


# if running pyLDT from conda environment disable compilation to avoid conflicts
import sys, os 
is_conda = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))
if is_conda:
    from julia.api import Julia
    jl = Julia(compiled_modules=False)

import numpy as np
import scipy.optimize
import scipy.integrate
import scipy.interpolate
import mcfit

from . import growth_eqns
from . import solve_eqns
from . import compute_pk


def init_ode():
    """
    Initialise all ODEs when loading pyLDT. Add here new ODE to initialise and increase N_models accordingly
    """

    N_models = 3
    res = [None for _ in range(N_models)]
    res[0] = solve_eqns.SphEvoGR(0.,0.3,1e-5,0.)
    res[1] = solve_eqns.calc_growth_fr_full(0.3,-1e-5,1.,0.001,0.)
    res[2] = solve_eqns.calc_growth_dgp_full(0.3,0.5,0.)

    for i,out in enumerate(res):
        if i==0 and np.isnan(out):
            print('ODE initialisation {:d} failed. Check growth_eqns.py and/or solve_eqns.py modules for bugs and dependencies.'.format(i))
        elif i>0 and np.any(np.isnan(out.u)):
            print('ODE initialisation {:d} failed. Check growth_eqns.py and/or solve_eqns.py modules for bugs and dependencies.'.format(i))

init_ode()


class matter_pdf:

    def __init__(self, params, zvec, Rvec, s2_mu_fid):
        """
        Create new instance of matter_pdf class with user-defined fiducial cosmology, redshifts, smoothing radii
        """

        self.z = np.asarray(zvec, dtype=np.float64)
        self.R = np.asarray(Rvec, dtype=np.float64)

        Nz = len(zvec)

        # To get P(k) at all redshifts rescale CAMB output at z=0 with LCDM growth (neglects perturbative and background effect from radiation)
        # This approach is more self-consistent with MG rescaling below, but could be avoided if using MGCAMB/EFTCAMB/Hi_CLASS to get P(k) in MG instead
        k_camb, _, pk_camb = compute_pk.compute_pk_gr(params,[0.])
        D0_lambda = scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,1.,args=(params['Omega_m'],1e-10,1.),tol=1e-8,rtol=1e-8)[0]
        Dz_lambda_norm = np.array([scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,1/(1+zf),args=(params['Omega_m'],1e-10,1/(1+zf)),tol=1e-8,rtol=1e-8)[0] for zf in zvec])/D0_lambda
        Dz_reshaped = Dz_lambda_norm.reshape(len(Dz_lambda_norm),1)
        pk_camb_z = np.repeat(Dz_reshaped**2, len(pk_camb[0]), axis=1) * np.repeat(pk_camb, Nz, axis=0)
        logpk = [scipy.interpolate.CubicSpline(np.log(k_camb),np.log(pk_camb_z[i])) for i in range(Nz)]
        #####################################

        def pk(k,z_ix):
            return np.exp(logpk[z_ix](np.log(k)))

        k_sample = np.logspace(-4,np.log10(20),num=512)
        self.sigma2_fid = self.compute_sigma2(k_sample,Nz,pk)

        self.s2_mu_fid = s2_mu_fid
        self.tau = self.get_tau(1., zf=0) # fix spherical evolution to EdS; sensitivity to DE/MG is negligible anyway


    def get_tau(self, Omega_m, zf=0):
        """
        Returns spline object to the linear density field tau as a function of the non-linear density rho computed in LCDM. Takes
        background total matter density and final redshift as input parameters.
        """

        # Integration interval
        ai = 1e-5
        af = 1/(1+zf)
        xi = np.log(ai)
        xf = np.log(af)

        N_dlin = 100
        N_dlin1 = 50
        N_dlin2 = N_dlin - N_dlin1 
        dlin_min = -5.5
        dlin_max = 1.3
        dlin_vec1 = np.linspace(dlin_min,0,N_dlin1,endpoint=False)
        dlin_vec2 = np.linspace(1e-4,dlin_max,N_dlin2)
        dlin_vec = np.concatenate((dlin_vec1,dlin_vec2))

        D_lambda = scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,af,args=(Omega_m,ai,af),tol=1e-8,rtol=1e-8)[0]
        dini_vec = dlin_vec/D_lambda

        return solve_eqns.calc_tau_gr(dini_vec, dlin_vec, Omega_m, xi, xf)


    def compute_sigma2(self, kvec, Nz, pk):
        """
        For a given linear power spectrum P(k,z), returns spline object to sigma^2(R) computed 
        with the Hankel Transform.
        """

        sigma2 = [None for _ in range(Nz)]
        for z_ix in range(Nz):
            
            def f1(k): return 9.*pk(k,z_ix)/(4.*np.pi*k*k)

            F = f1(kvec)
            DBT = mcfit.DoubleBessel(kvec,alpha=1,nu=1.5,q=-1,lowring=True)
            Rvec, G = DBT(F,extrap=True)
            sigma2[z_ix] = scipy.interpolate.CubicSpline(Rvec,G/Rvec**3.)

        return sigma2


    def get_s2_mu(self, sigma2):
        """
        Uses the log-normal approximation to get non-linear variance of log-density field from 
        that of the fiducial input cosmology
        """

        s2_mu = np.zeros_like(self.s2_mu_fid)

        for ix,rth in enumerate(self.R):
            sigma2_all_z = np.asarray([sig2(rth) for sig2 in sigma2])
            sigma2_fid_all_z = np.asarray([sig2(rth) for sig2 in self.sigma2_fid])
            s2_mu[ix] = np.log(1 + sigma2_all_z)/np.log(1 + sigma2_fid_all_z) * self.s2_mu_fid[ix]
        
        return s2_mu
    

    def sigma2_gr(self, params):
        """
        Compute a list of length Nz of sigma_lin^2(R,z[i]) functions for GR given an input cosmology and 
        also returns the z=0 GR linear matter power spectrum and wavenumbers. 
        """

        Nz = len(self.z)
        # To get P(k) at all redshifts rescale CAMB output at z=0 with LCDM growth (i.e. neglects perturbative and background effect from radiation)
        # This approach is more self-consistent with MG rescaling below, but could be avoided if using MGCAMB/EFTCAMB/Hi_CLASS to get P(k) in MG instead
        k_camb, _, pk_camb = compute_pk.compute_pk_gr(params,[0.])
        D0_lambda = scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,1.,args=(params['Omega_m'],1e-10,1.),tol=1e-8,rtol=1e-8)[0]
        Dz_lambda_norm = np.array([scipy.integrate.quadrature(growth_eqns.growth_int,1e-10,1/(1+zf),args=(params['Omega_m'],1e-10,1/(1+zf)),tol=1e-8,rtol=1e-8)[0] for zf in self.z])/D0_lambda
        Dz_reshaped = Dz_lambda_norm.reshape(len(Dz_lambda_norm),1)
        pk_camb_z = np.repeat(Dz_reshaped**2, len(pk_camb[0]), axis=1) * np.repeat(pk_camb, Nz, axis=0)
        logpk = [scipy.interpolate.CubicSpline(np.log(k_camb),np.log(pk_camb_z[i])) for i in range(Nz)]

        def pk(k,z_ix):
            return np.exp(logpk[z_ix](np.log(k)))

        k_sample = np.logspace(-4,np.log10(20),num=512)
        sigma2 = self.compute_sigma2(k_sample,Nz,pk)

        return sigma2, k_camb, pk_camb


    def sigma2_fr(self, params, kvec, pk_gr):
        """
        Compute a list of length Nz of sigma_lin^2(R,z[i]) functions for f(R) gravity given an input cosmology 
        and GR linear power spectrum at z=0
        """

        Nz = len(self.z)

        pk_vec = compute_pk.compute_pk_fr_vectorized(params['Omega_m'], params['fR0'], params['n'], self.z, kvec, pk_gr)
        logpk = [scipy.interpolate.CubicSpline(np.log(kvec),np.log(pk_vec[i])) for i in range(Nz)]

        def pk(k,z_ix):
            return np.exp(logpk[z_ix](np.log(k)))

        k_sample = np.logspace(-4,np.log10(20),num=512)
        sigma2 = self.compute_sigma2(k_sample,Nz,pk)

        return sigma2


    def sigma2_dgp(self, params, kvec, pk_gr):
        """
        Compute a list of length Nz of sigma_lin^2(R,z[i]) functions for nDGP gravity given an input cosmology 
        and GR linear power spectrum at z=0
        """

        Nz = len(self.z)

        pk_vec = compute_pk.compute_pk_dgp_vectorized(params['Omega_m'], params['rcH0'], self.z, pk_gr)
        logpk = [scipy.interpolate.CubicSpline(np.log(kvec),np.log(pk_vec[i])) for i in range(Nz)]

        def pk(k,z_ix):
            return np.exp(logpk[z_ix](np.log(k)))

        k_sample = np.logspace(-4,np.log10(20),num=512)
        sigma2 = self.compute_sigma2(k_sample,Nz,pk)

        return sigma2


    def get_pdf(self, s2_mu, sigma2):
        """
        Return matter PDF for a given matrix [i,j] of non-linear log-density variances for smoothing radii R[i] and 
        redshifts z[j] as well as a list of sigma_lin^2(R,z[i]) functions.
        """

        z = self.z
        R = self.R
        tau = self.tau
        Nz = len(z)
        Nr = len(R)

        rho_vec = np.linspace(0.1,13.5,1000)

        def get_psi(R, s2_mu, sigma2):
        # Here R, s2_mu and sigma2 refer to particular elements in the corresponding input arrays
            psi_vec = (sigma2(R)/s2_mu) * (tau(rho_vec)**2/(2*sigma2(R*rho_vec**(1/3))))
            return scipy.interpolate.CubicSpline(rho_vec, psi_vec)

        psi_mat = [[get_psi(rth, s2_mu[r_ix,z_ix], sigma2[z_ix]) for z_ix in range(Nz)] for r_ix,rth in enumerate(R)]

        def get_pdf_nn(psi):
        # compute non-normalised PDF
            pdf_nn_vec = np.sqrt((psi(rho_vec,2) + psi(rho_vec,1)/rho_vec)/(2*np.pi))*np.exp(-psi(rho_vec))
            return scipy.interpolate.CubicSpline(rho_vec, pdf_nn_vec)

        pdf_nn_mat = [[get_pdf_nn(psi_mat[r_ix][z_ix]) for z_ix in range(Nz)] for r_ix in range(Nr)]

        def mean_igr(rho, pdf_nn):
        # non-normalised PDF mean integrand
            return rho * pdf_nn(rho)

        mean_mat = [[scipy.integrate.quad(mean_igr,0.1,10,args=(pdf_nn_mat[r_ix][z_ix]),epsabs=1e-4,epsrel=1e-4,limit=50)[0] for z_ix in range(Nz)] for r_ix in range(Nr)]
        norm_mat = [[scipy.integrate.quad(pdf_nn_mat[r_ix][z_ix],0.1,10,epsabs=1e-4,epsrel=1e-4,limit=50)[0] for z_ix in range(Nz)] for r_ix in range(Nr)]

        def pdf(pdf_nn,mean,norm):
            rho_max = 10.
            pdf_vec = pdf_nn(rho_vec[rho_vec<rho_max] * mean/norm) * mean/norm**2
            return scipy.interpolate.CubicSpline(rho_vec[rho_vec<rho_max], pdf_vec)

        pdf_mat = [[pdf(pdf_nn_mat[r_ix][z_ix],mean_mat[r_ix][z_ix],norm_mat[r_ix][z_ix]) for z_ix in range(Nz)] for r_ix in range(Nr)]

        return pdf_mat


    def compute_pdf(self, cosmo_params, models=['gr']):
        """
        Compute matter PDF for requested cosmo_params and models. The output is a dictionary with keys 'gr' and/or 'fr' and/or 'dgp'. 
        Each value is a matrix [i][j] where [i] runs over (increasing) smoothing radius and [j] over (increasing) redshifts.
        """

        # compute linear and non-linear variances
        sig2_gr, k_camb, pk_camb = self.sigma2_gr(cosmo_params)
        s2_mu_gr = self.get_s2_mu(sig2_gr)
        if 'fr' in models: 
            sig2_fr = self.sigma2_fr(cosmo_params, k_camb, pk_camb)
            s2_mu_fr = self.get_s2_mu(sig2_fr)
        if 'dgp' in models: 
            sig2_dgp = self.sigma2_dgp(cosmo_params, k_camb, pk_camb)
            s2_mu_dgp = self.get_s2_mu(sig2_dgp)

        # compute matter PDF
        pdf_dict = {}
        if 'gr' in models:
            pdf_gr_mat = self.get_pdf(s2_mu_gr, sig2_gr)
            pdf_dict['gr'] = pdf_gr_mat
        if 'fr' in models: 
            pdf_fr_mat = self.get_pdf(s2_mu_fr, sig2_fr)
            pdf_dict['fr'] = pdf_fr_mat
        if 'dgp' in models: 
            pdf_dgp_mat = self.get_pdf(s2_mu_dgp, sig2_dgp)
            pdf_dict['dgp'] = pdf_dgp_mat
        
        return pdf_dict

    ######################### Test functions below #########################

    


from julia import Main

# Top-hat evolution equation in flat GR+LCDM; for \rho = (1 + delta_NL) in [0.1, 10], 
# spherical evolution in MG with LCDM expansion follows by simply rescaling the 
# initial conditions with linear theory ratios
jul_rhs_gr = Main.eval("""
    function rhs_gr(dy,y,p,x)
        Omega_m, deltai, xi = p
        Omega_L = 1. - Omega_m
        H = sqrt(Omega_m * exp(-3*x) + Omega_L)
        H2 = Omega_m * exp(-3*x) + Omega_L
        dH = -(3*exp(-3*x) * Omega_m)/(2 * H)
        delta_NL = (1. + deltai) / (exp(xi-x)*y[1] + 1.)^3 - 1.
    
        dy[1] = y[2]
        dy[2] = -(dH/H)*y[2] - 0.5 * ((Omega_m * exp(-3*x) - 2. * Omega_L)/H2) * y[1] - 0.5 * Omega_m * exp(-3*x) / H2 * (exp(x-xi) + y[1]) * delta_NL
    end
    """)

# add you favourite spherical evolution here...


import setuptools
import os
import re

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(os.path.join(os.path.dirname(__file__), 'pyLDT_cosmo', '__init__.py')) as fp:
    for line in fp:
        m = re.search(r'^\s*__version__\s*=\s*([\'"])([^\'"]+)\1\s*$', line)
        if m:
            version = m.group(2)
            break
    else:
        raise RuntimeError('Unable to find own __version__ string')

setuptools.setup(
    name="pyLDT-cosmo",
    version=version,
    author="Matteo Cataneo",
    author_email="mcataneo85@gmail.com",
    description="Package for PDF calculations in Large Deviation Theory",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mcataneo/pyLDT-cosmo",
    packages=setuptools.find_packages(),
    package_data={'pyLDT_cosmo': ['benchmarks/*.dat']},
    install_requires=[
        'numpy>=1.19',
        'scipy>=1.5',
        'camb==1.1.2',
        'mcfit==0.0.16',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
